# -*- coding: utf-8 -*-
"""This module provides :class:`.KDTree` class as an interface to Thomas
Hamelryck's KDTree C module distributed with Biopython."""

from .kdtree import KDTree

__all__ = ['KDTree']
