.subnav {
    width: 100%;
    height: 36px;
    background-color: #eeeeee; /* Old browsers */
    background-repeat: repeat-x; /* Repeat the gradient */
    background-image: -moz-linear-gradient(top, #f5f5f5 0%, #eeeeee 100%); /* FF3.6+ */
    background-image: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#f5f5f5), color-stop(100%,#eeeeee)); /* Chrome,Safari4+ */
    background-image: -webkit-linear-gradient(top, #f5f5f5 0%,#eeeeee 100%); /* Chrome 10+,Safari 5.1+ */
    background-image: -ms-linear-gradient(top, #f5f5f5 0%,#eeeeee 100%); /* IE10+ */
    background-image: -o-linear-gradient(top, #f5f5f5 0%,#eeeeee 100%); /* Opera 11.10+ */
    filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#f5f5f5', endColorstr='#eeeeee',GradientType=0 ); /* IE6-9 */
    background-image: linear-gradient(top, #f5f5f5 0%,#eeeeee 100%); /* W3C */
    border: 1px solid #e5e5e5;
    -webkit-border-radius: 4px;
    -moz-border-radius: 4px;
    border-radius: 4px;
    margin-top: 10px;
}

.subnav .nav {
    margin-bottom: 0;
}

.subnav .nav > li > a {
    margin: 0;
    padding-top:    11px;
    padding-bottom: 11px;
    border-left: 1px solid #f5f5f5;
    border-right: 1px solid #e5e5e5;
    -webkit-border-radius: 0;
    -moz-border-radius: 0;
    border-radius: 0;
}

.subnav .nav > .active > a,
.subnav .nav > .active > a:hover {
    padding-left: 13px;
    color: #777;
    background-color: #e9e9e9;
    border-right-color: #ddd;
    border-left: 0;
    -webkit-box-shadow: inset 0 3px 5px rgba(0,0,0,.05);
    -moz-box-shadow: inset 0 3px 5px rgba(0,0,0,.05);
    box-shadow: inset 0 3px 5px rgba(0,0,0,.05);
}

.subnav .nav > .active > a .caret,
.subnav .nav > .active > a:hover .caret {
    border-top-color: #777;
}

.subnav .nav > li:first-child > a,
.subnav .nav > li:first-child > a:hover {
    border-left: 0;
    padding-left: 12px;
    -webkit-border-radius: 4px 0 0 4px;
    -moz-border-radius: 4px 0 0 4px;
    border-radius: 4px 0 0 4px;
}

.subnav .nav > li:last-child > a {
    border-right: 0;
}

.subnav .dropdown-menu {
    -webkit-border-radius: 0 0 4px 4px;
    -moz-border-radius: 0 0 4px 4px;
    border-radius: 0 0 4px 4px;
}

.subnav.affix {
    position: fixed;
    margin-top: 0px;
    top: 0px;
    left: 0;
    right: 0;
    z-index: 1020; /* 10 less than .navbar-fixed to prevent any overlap */
    border-color: #d5d5d5;
    border-width: 0 0 1px; /* drop the border on the fixed edges */
    -webkit-border-radius: 0;
    -moz-border-radius: 0;
    border-radius: 0;
    -webkit-box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    -moz-box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    filter: progid:DXImageTransform.Microsoft.gradient(enabled=false); /* IE6-9 */
    opacity: 0.0;
    -webkit-transition:opacity 1s ease-in-out;
    -moz-transition:opacity 1s ease-in-out;
    -o-transition:opacity 1s ease-in-out;
    -ms-transition:opacity 1s ease-in-out;
    transition:opacity 1s ease-in-out;
}

.subnav.affix:hover {
    position: fixed;
    margin-top: 0px;
    top: 0px;
    left: 0;
    right: 0;
    z-index: 1020; /* 10 less than .navbar-fixed to prevent any overlap */
    border-color: #d5d5d5;
    border-width: 0 0 1px; /* drop the border on the fixed edges */
    -webkit-border-radius: 0;
    -moz-border-radius: 0;
    border-radius: 0;
    -webkit-box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    -moz-box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    box-shadow: inset 0 1px 0 #fff, 0 1px 5px rgba(0,0,0,.1);
    filter: progid:DXImageTransform.Microsoft.gradient(enabled=false); /* IE6-9 */
    opacity: 1;
    -webkit-transition:opacity 0.2s ease-in-out;
    -moz-transition:opacity 0.2s ease-in-out;
    -o-transition:opacity 0.2s ease-in-out;
    -ms-transition:opacity 0.2s ease-in-out;
    transition:opacity 0.2s ease-in-out;
}

.subnav.affix .nav {
    width: 938px;
    margin: 0 auto;
    padding: 0 1px;
}

.subnav .nav > li:first-child > a,
.subnav .nav > li:first-child > a:hover {
    -webkit-border-radius: 0;
    -moz-border-radius: 0;
    border-radius: 0;
}

.subnav.affix {
  top: 0px;
}


/* ToC
------------------------- */

.toc ul {
  list-style-type: none;
  padding: 0;
  margin-left: 0;
}
.toc > li > ul > li > a {
  padding-left: 25px;
}
.toc > li > ul > li > ul > li > a {
  padding-left: 40px;
}
.toc > li > ul > li > ul > li > ul > li > a {
  padding-left: 55px;
}

.toc {
  width: 220px;
  margin: 0 0 0 0;
  padding: 0;
  background-color: #fff;
  -webkit-border-radius: 6px;
     -moz-border-radius: 6px;
          border-radius: 6px;
  -webkit-box-shadow: 0 1px 4px rgba(0,0,0,.065);
     -moz-box-shadow: 0 1px 4px rgba(0,0,0,.065);
          box-shadow: 0 1px 4px rgba(0,0,0,.065);
}
.toc li > a {
  display: block;
  width: 190px \9;
  margin: 0 0 -1px;
  border: 1px solid #e5e5e5;
  -webkit-box-shadow: 0 1px 4px rgba(0,0,0,.065);
     -moz-box-shadow: 0 1px 4px rgba(0,0,0,.065);
          box-shadow: 0 1px 4px rgba(0,0,0,.065);
}
.toc li > a.navfirst {
  -webkit-border-top-left-radius: 6px;
     -moz-border-top-left-radius: 6px;
          border-top-left-radius: 6px;
  -webkit-border-top-right-radius: 6px;
     -moz-border-top-right-radius: 6px;
          border-top-right-radius: 6px;
}
.toc li > a.navlast {
  -webkit-border-bottom-left-radius: 6px;
     -moz-border-bottom-left-radius: 6px;
          border-bottom-left-radius: 6px;
  -webkit-border-bottom-right-radius: 6px;
     -moz-border-bottom-right-radius: 6px;
          border-bottom-right-radius: 6px;
}
.toc .active > a {
  position: relative;
  z-index: 2;
  border: 0;
  text-shadow: 0 1px 0 rgba(0,0,0,.15);
  -webkit-box-shadow: inset 1px 0 0 rgba(0,0,0,.1), inset -1px 0 0 rgba(0,0,0,.1);
     -moz-box-shadow: inset 1px 0 0 rgba(0,0,0,.1), inset -1px 0 0 rgba(0,0,0,.1);
          box-shadow: inset 1px 0 0 rgba(0,0,0,.1), inset -1px 0 0 rgba(0,0,0,.1);
}
/* Chevrons */
.toc .icon-chevron-right {
  float: right;
  margin-top: 2px;
  margin-right: -6px;
  opacity: .25;
}
.toc > li > a:hover {
  background-color: #f5f5f5;
}
.toc a:hover .icon-chevron-right {
  opacity: .5;
}
.toc .active > a > .icon-chevron-right,
.toc .active > a:hover .icon-chevron-right {
  background-image: url(../img/glyphicons-halflings-white.png);
  opacity: 1;
}

#sidebar {
  padding-top: 20px;
}
#sidebar.affix {
  top: 30px;
}
#sidebar.affix-bottom {
  position: absolute;
  top: auto;
  bottom: 270px;
}

/* Source
------------------------- */

.viewcode-link {
    float: right;
}

.viewcode-back {
    float: right;
    font-family: sans-serif;
}

div.viewcode-block:target {
    margin: -1px -10px;
    padding: 0 10px;
}

.viewcode-back {
    font-family: 'Lucida Grande', 'Lucida Sans Unicode', 'Geneva',
                 'Verdana', sans-serif;
    float: right;
}

div.viewcode-block:target {
    background-color: #f4debf;
    border-top: 1px solid #ac9;
    border-bottom: 1px solid #ac9;
    margin: -1px -10px;
    padding: 0 10px;
}

dl.class, dl.function {
    border-top: 2px solid #888;
}

dl.method, dl.attribute {
    border-top: 1px solid #aaa;
}

div#content {
    position: relative;
    top: -100px;
}



/* Footer
-------------------------------------------------- */

.footer {
  text-align: right;
  padding: 30px 0;
  margin-top: 40px;
  border-top: 1px solid #e5e5e5;
}
.footer p {
  margin-bottom: 0;
  color: #777;
}



pre {
    overflow: auto;
    overflow-y: hidden;  /* fixes display issues on Chrome browsers */
}

/* Sections
-------------------------------------------------- */

/* padding for in-page bookmarks and fixed navbar */
section {
  padding-top: 10px;
}

.section {
  padding-top: 10px;
}
.section > .page-header,
.section > .lead {
  color: #5a5a5a;
}
.section > ul li {
  margin-bottom: 5px;
}

/* Separators (hr) */
.bs-docs-separator {
  margin: 40px 0 39px;
}

.modal#codemodal {
  width: 800px;
  margin-left: -400px;
}

#codesnippets {
  width: 755px;
  height: 360px;
  resize: none;
  font-size: 9pt;
  font-family: Consolas,Monaco,Lucida Console,Liberation Mono,DejaVu Sans Mono,Bitstream Vera Sans Mono,Courier New, monospace;
}

td.label {
  display: inline-block;
  padding: 2px 4px;
  font-size: inherit;
  font-weight: normal;
  line-height: 14px;
  color: inherit;
  text-shadow: 0 0;
  white-space: nowrap;
  vertical-align: bottom;
  background-color: #fff;
}

tt {
  font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
  padding: 2px 4px;
  color: #d14;
  white-space: nowrap;
  background-color: #f7f7f9;
  border: 1px solid #e1e1e8;
}

a > tt {
  color: #99356b;
}

.guilabel, .menuselection {
  padding-right: 9px;
  padding-left: 9px;
  -webkit-border-radius: 9px;
  -moz-border-radius: 9px;
  border-radius: 9px;
  background-color: #3a87ad;
  display: inline-block;
  padding: 2px 4px;
  font-size: 11.844px;
  font-weight: bold;
  line-height: 14px;
  color: #fff;
  text-shadow: 0 -1px 0 rgba(0,0,0,0.25);
  white-space: nowrap;
  vertical-align: baseline;
}