/*
*
* Template Numerical Toolkit (TNT)
*
* Mathematical and Computational Sciences Division
* National Institute of Technology,
* Gaithersburg, MD USA
*
*
* This software was developed at the National Institute of Standards and
* Technology (NIST) by employees of the Federal Government in the course
* of their official duties. Pursuant to title 17 Section 105 of the
* United States Code, this software is not subject to copyright protection
* and is in the public domain. NIST assumes no responsibility whatsoever for
* its use by other parties, and makes no guarantees, expressed or implied,
* about its quality, reliability, or any other characteristic.
*
*/

#ifndef TNT_VERSION_H
#define TNT_VERSION_H


//---------------------------------------------------------------------
//  current version 
//---------------------------------------------------------------------


#define TNT_MAJOR_VERSION    '1'
#define TNT_MINOR_VERSION    '2'
#define TNT_SUBMINOR_VERSION '6'
#define TNT_VERSION_STRING "1.2.6"





#endif
// TNT_VERSION_H
